import syft as sy

import torch as th
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from syft.serde.serde import deserialize
from syft.serde.serde import serialize
from syft import messaging
import time

import pytest
import unittest.mock as mock

from syft.workers import WebsocketClientWorker
from syft.workers import WebsocketServerWorker


def test_plan_built_automatically(hook):
    # To run a plan locally the local worker can't be a client worker,
    # since it needs to register objects
    hook.local_worker.is_client_worker = False

    @sy.func2plan(args_shape=[(1,)])
    def plan_abs(data):
        return data.abs()

    assert isinstance(plan_abs.__str__(), str)
    assert len(plan_abs.readable_plan) > 0
    assert plan_abs.is_built

    hook.local_worker.is_client_worker = True


def test_plan_build(hook):
    hook.local_worker.is_client_worker = False

    @sy.func2plan(args_shape=())
    def plan_abs(data):
        return data.abs()

    assert not plan_abs.is_built
    assert not len(plan_abs.readable_plan)

    plan_abs.build(th.tensor([-1]))

    assert len(plan_abs.readable_plan)
    assert plan_abs.is_built

    hook.local_worker.is_client_worker = True


def test_plan_built_automatically_with_any_dimension(hook):
    hook.local_worker.is_client_worker = False

    @sy.func2plan(args_shape=[(-1, 1)])
    def plan_abs(data):
        return data.abs()

    assert isinstance(plan_abs.__str__(), str)
    assert len(plan_abs.readable_plan) > 0

    hook.local_worker.is_client_worker = True


def test_raise_exception_for_invalid_shape(hook):
    hook.local_worker.is_client_worker = False

    with pytest.raises(ValueError):

        @sy.func2plan(args_shape=[(1, -20)])
        def _(data):
            return data  # pragma: no cover

    hook.local_worker.is_client_worker = True


def test_raise_exception_when_sending_unbuilt_plan(workers):
    me = workers["me"]
    me.is_client_worker = False

    bob = workers["bob"]

    @sy.func2plan()
    def plan(data):
        return data  # pragma: no cover

    with pytest.raises(RuntimeError):
        plan.send(bob)

    me.is_client_worker = True


def test_plan_execute_locally(hook):
    hook.local_worker.is_client_worker = False

    @sy.func2plan(args_shape=[(1,)])
    def plan_abs(data):
        return data.abs()

    x = th.tensor([-1, 2, 3])
    x_abs = plan_abs(x)
    assert (x_abs == th.tensor([1, 2, 3])).all()

    hook.local_worker.is_client_worker = True


def test_plan_method_execute_locally(hook):
    hook.local_worker.is_client_worker = False

    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = nn.Linear(2, 3)
            self.fc2 = nn.Linear(3, 2)
            self.fc3 = nn.Linear(2, 1)

        @sy.method2plan
        def forward(self, x):
            x = F.relu(self.fc1(x))
            x = self.fc2(x)
            x = self.fc3(x)
            return F.log_softmax(x)

    model = Net()

    # Force build
    assert model(th.tensor([1.0, 2])) == 0

    # Test call multiple times
    assert model(th.tensor([1.0, 2.1])) == 0

    hook.local_worker.is_client_worker = True


def test_plan_multiple_send(workers):
    me = workers["me"]
    me.is_client_worker = False

    bob = workers["bob"]
    alice = workers["alice"]

    @sy.func2plan(args_shape=[(1,)])
    def plan_abs(data):
        return data.abs()

    plan_abs.send(bob)
    x_ptr = th.tensor([-1, 7, 3]).send(bob)
    p = plan_abs(x_ptr)
    x_abs = p.get()

    assert (x_abs == th.tensor([1, 7, 3])).all()

    # Test get / send plan
    plan_abs.get()
    plan_abs.send(alice)

    x_ptr = th.tensor([-1, 2, 3]).send(alice)
    p = plan_abs(x_ptr)
    x_abs = p.get()
    assert (x_abs == th.tensor([1, 2, 3])).all()
    me.is_client_worker = True


def test_plan_built_on_method(hook):
    """
    Test @sy.meth2plan and plan send / get / send
    """
    hook.local_worker.is_client_worker = False

    x11 = th.tensor([-1, 2.0]).tag("input_data")
    x21 = th.tensor([-1, 2.0]).tag("input_data")

    device_1 = sy.VirtualWorker(hook, id="device_1", data=(x11,))
    device_2 = sy.VirtualWorker(hook, id="device_2", data=(x21,))

    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = nn.Linear(2, 3)
            self.fc2 = nn.Linear(3, 2)

        @sy.method2plan
        def forward(self, x):
            x = F.relu(self.fc1(x))
            x = self.fc2(x)
            return F.log_softmax(x, dim=0)

    net = Net()

    # build
    net.forward.build(th.tensor([1, 2.0]))

    net.send(device_1)
    pointer_to_data = device_1.search("input_data")[0]
    pointer_to_result = net(pointer_to_data)

    assert isinstance(pointer_to_result.get(), th.Tensor)

    net.get()
    net.send(device_2)

    pointer_to_data = device_2.search("input_data")[0]
    pointer_to_result = net(pointer_to_data)

    assert isinstance(pointer_to_result.get(), th.Tensor)

    hook.local_worker.is_client_worker = True


def test_multiple_workers(workers):
    me = workers["me"]
    me.is_client_worker = False
    bob = workers["bob"]
    alice = workers["alice"]

    @sy.func2plan(args_shape=[(1,)])
    def plan_abs(data):
        return data.abs()

    plan_abs.send(bob, alice)
    x_ptr = th.tensor([-1, 7, 3]).send(bob)
    p = plan_abs(x_ptr)
    x_abs = p.get()
    assert (x_abs == th.tensor([1, 7, 3])).all()

    x_ptr = th.tensor([-1, 9, 3]).send(alice)
    p = plan_abs(x_ptr)
    x_abs = p.get()
    assert (x_abs == th.tensor([1, 9, 3])).all()

    me.is_client_worker = True


def test_fetch_plan(hook):
    hook.local_worker.is_client_worker = False
    device_4 = sy.VirtualWorker(hook, id="device_4")

    @sy.func2plan(args_shape=[(1,)])
    def plan_mult_3(data):
        return data * 3

    sent_plan = plan_mult_3.send(device_4)

    # Fetch plan
    fetched_plan = device_4.fetch_plan(sent_plan.id)
    get_plan = sent_plan.get()

    # Execut it locally
    x = th.tensor([-1, 2, 3])
    assert (get_plan(x) == th.tensor([-3, 6, 9])).all()
    assert fetched_plan.is_built
    assert (fetched_plan(x) == th.tensor([-3, 6, 9])).all()

    hook.local_worker.is_client_worker = True


def test_plan_serde(hook):
    hook.local_worker.is_client_worker = False

    @sy.func2plan(args_shape=[(1, 3)])
    def my_plan(data):
        x = data * 2
        y = (x - 2) * 10
        return x + y

    serialized_plan = serialize(my_plan)
    deserialized_plan = deserialize(serialized_plan)

    x = th.tensor([-1, 2, 3])
    assert (deserialized_plan(x) == th.tensor([-42, 24, 46])).all()

    hook.local_worker.is_client_worker = True


def test_execute_plan_remotely(hook, start_proc):
    """Test plan execution remotely."""
    hook.local_worker.is_client_worker = False

    @sy.func2plan(args_shape=[(1,)])
    def my_plan(data):
        x = data * 2
        y = (x - 2) * 10
        return x + y

    x = th.tensor([-1, 2, 3])
    local_res = my_plan(x)

    kwargs = {"id": "test_plan_worker", "host": "localhost", "port": 8799, "hook": hook}
    server = start_proc(WebsocketServerWorker, **kwargs)

    time.sleep(0.1)
    socket_pipe = WebsocketClientWorker(**kwargs)

    plan_ptr = my_plan.send(socket_pipe)
    x_ptr = x.send(socket_pipe)
    plan_res = plan_ptr(x_ptr).get()

    assert (plan_res == local_res).all()

    # delete remote object before websocket connection termination
    del x_ptr

    server.terminate()
    hook.local_worker.is_client_worker = True


def test_execute_plan_module_remotely(hook, start_proc):
    """Test plan execution remotely."""
    hook.local_worker.is_client_worker = False

    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = nn.Linear(2, 3)
            self.fc2 = nn.Linear(3, 2)

        @sy.method2plan
        def forward(self, x):
            x = F.relu(self.fc1(x))
            x = self.fc2(x)
            return F.log_softmax(x, dim=0)

    net = Net()

    x = th.tensor([-1, 2.0])
    local_res = net(x)
    assert not net.forward.is_built

    net.forward.build(x)

    kwargs = {"id": "test_plan_worker_2", "host": "localhost", "port": 8799, "hook": hook}
    server = start_proc(WebsocketServerWorker, **kwargs)

    time.sleep(0.1)
    socket_pipe = WebsocketClientWorker(**kwargs)

    plan_ptr = net.send(socket_pipe)
    x_ptr = x.send(socket_pipe)
    remote_res = plan_ptr(x_ptr).get()

    assert (remote_res == local_res).all()

    # delete remote object before websocket connection termination
    del x_ptr

    server.terminate()
    hook.local_worker.is_client_worker = True


def test_train_plan_locally_and_then_send_it(hook, start_proc):
    """Test training a plan locally and then executing it remotely."""
    hook.local_worker.is_client_worker = False

    # Create toy model
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = nn.Linear(2, 3)
            self.fc2 = nn.Linear(3, 2)

        @sy.method2plan
        def forward(self, x):
            x = F.relu(self.fc1(x))
            x = self.fc2(x)
            return F.log_softmax(x, dim=0)

    net = Net()

    # Create toy data
    x = th.tensor([-1, 2.0])
    y = th.tensor([1.0])

    # Train Model
    opt = optim.SGD(params=net.parameters(), lr=0.01)
    previous_loss = None

    for _ in range(5):
        # 1) erase previous gradients (if they exist)
        opt.zero_grad()

        # 2) make a prediction
        pred = net(x)

        # 3) calculate how much we missed
        loss = ((pred - y) ** 2).sum()

        # 4) figure out which weights caused us to miss
        loss.backward()

        # 5) change those weights
        opt.step()

        if previous_loss is not None:
            assert loss < previous_loss

        previous_loss = loss

    local_res = net(x)
    net.forward.build(x)

    kwargs = {"id": "test_plan_worker_3", "host": "localhost", "port": 8800, "hook": hook}
    server = start_proc(WebsocketServerWorker, **kwargs)

    time.sleep(0.1)
    socket_pipe = WebsocketClientWorker(**kwargs)

    plan_ptr = net.send(socket_pipe)
    x_ptr = x.send(socket_pipe)
    remote_res = plan_ptr(x_ptr).get()

    assert (remote_res == local_res).all()

    # delete remote object before websocket connection termination
    del x_ptr

    server.terminate()
    hook.local_worker.is_client_worker = True


def test_replace_worker_ids_two_strings(hook):
    plan = sy.Plan(id="0", owner=hook.local_worker, name="test_plan")
    _replace_message_ids_orig = messaging.Plan._replace_message_ids
    mock_fun = mock.Mock(return_value=[])
    messaging.Plan._replace_message_ids = mock_fun
    plan.replace_worker_ids("me", "you")
    args = {"change_id": -1, "obj": [], "to_id": -1}
    calls = [
        mock.call(from_worker="me", to_worker="you", **args),
        mock.call(from_worker=b"me", to_worker=b"you", **args),
    ]
    assert len(mock_fun.mock_calls) == 2
    mock_fun.assert_has_calls(calls, any_order=True)
    messaging.Plan._replace_message_ids = _replace_message_ids_orig


def test_replace_worker_ids_one_string_one_int(hook):
    plan = sy.Plan(id="0", owner=hook.local_worker, name="test_plan")
    _replace_message_ids_orig = messaging.Plan._replace_message_ids

    mock_fun = mock.Mock(return_value=[])
    messaging.Plan._replace_message_ids = mock_fun
    plan.replace_worker_ids(100, "you")

    args = {"change_id": -1, "obj": [], "to_id": -1}
    calls = [mock.call(from_worker=100, to_worker="you", **args)]
    assert len(mock_fun.mock_calls) == 1
    mock_fun.assert_has_calls(calls, any_order=True)

    mock_fun = mock.Mock(return_value=[])
    messaging.Plan._replace_message_ids = mock_fun
    plan.replace_worker_ids("me", 200)
    calls = [
        mock.call(from_worker="me", to_worker=200, **args),
        mock.call(from_worker=b"me", to_worker=200, **args),
    ]
    assert len(mock_fun.mock_calls) == 2
    mock_fun.assert_has_calls(calls, any_order=True)
    messaging.Plan._replace_message_ids = _replace_message_ids_orig


def test_replace_worker_ids_two_ints(hook):
    plan = sy.Plan(id="0", owner=hook.local_worker, name="test_plan")
    _replace_message_ids_orig = messaging.Plan._replace_message_ids
    mock_fun = mock.Mock(return_value=[])
    messaging.Plan._replace_message_ids = mock_fun
    plan.replace_worker_ids(300, 400)
    args = {"change_id": -1, "obj": [], "to_id": -1}
    calls = [mock.call(from_worker=300, to_worker=400, **args)]
    mock_fun.assert_called_once()
    mock_fun.assert_has_calls(calls, any_order=True)
    messaging.Plan._replace_message_ids = _replace_message_ids_orig


def test__replace_message_ids():
    messages = [10, ("worker", "me"), "you", 20, 10, b"you", (30, ["you", "me", "bla"])]

    replaced = messaging.Plan._replace_message_ids(
        obj=messages, change_id=10, to_id=100, from_worker="me", to_worker="another"
    )

    # note that tuples are converted to lists
    expected = (100, ("worker", "another"), "you", 20, 100, b"you", (30, ("you", "another", "bla")))

    assert replaced == expected


def test___call__function(hook):
    plan = sy.Plan(id="0", owner=hook.local_worker, name="test_plan")
    with pytest.raises(ValueError):
        plan(kwarg1="hello", kwarg2=None)

    result_id = 444

    pop_function_original = sy.ID_PROVIDER.pop
    sy.ID_PROVIDER.pop = mock.Mock(return_value=result_id)

    return_value = "return value"
    plan.execute_plan = mock.Mock(return_value=return_value)

    arg_list = (100, 200, 356)
    ret_val = plan(*arg_list)

    plan.execute_plan.assert_called_with(arg_list, [result_id])

    assert ret_val == return_value

    # reset function
    sy.ID_PROVIDER.pop = pop_function_original


def test__call__raise(hook):
    plan = sy.Plan(id="0", owner=hook.local_worker, name="test_plan")
    with pytest.raises(ValueError):
        plan(kwarg1="hello", kwarg2=None)


def test__call__for_method(hook):
    plan = sy.Plan(id="0", owner=hook.local_worker, name="test_plan", is_method=True)
    result_id = 444

    pop_function_original = sy.ID_PROVIDER.pop
    sy.ID_PROVIDER.pop = mock.Mock(return_value=result_id)

    return_value = "return value"
    plan.execute_plan = mock.Mock(return_value=return_value)

    self_value = mock.Mock()
    self_value.send = mock.Mock()
    plan._self = self_value

    arg_list = (100, 200, 356)
    ret_val = plan(*arg_list)

    expected_args = tuple([self_value] + list(arg_list))
    plan.execute_plan.assert_called_with(expected_args, [result_id])

    assert ret_val == return_value

    # reset function
    sy.ID_PROVIDER.pop = pop_function_original



import pytest

import torch

import syft as sy
from syft import federated
from syft.frameworks.torch import pointers
from syft.frameworks.torch.federated import utils

PRINT_IN_UNITTESTS = False

# To make execution deterministic to some extent
# for more information - refer https://pytorch.org/docs/stable/notes/randomness.html
torch.manual_seed(0)
torch.cuda.manual_seed_all(0)
torch.backends.cudnn.benchmark = False
torch.backends.cudnn.deterministic = True


def test_add_dataset():
    fed_client = federated.FederatedClient()

    dataset = "my_dataset"
    fed_client.add_dataset(dataset, "string_dataset")

    assert "string_dataset" in fed_client.datasets


def test_add_dataset_with_duplicate_key():
    fed_client = federated.FederatedClient()

    dataset = "my_dataset"
    fed_client.add_dataset(dataset, "string_dataset")

    assert "string_dataset" in fed_client.datasets

    with pytest.raises(ValueError):
        fed_client.add_dataset(dataset, "string_dataset")


def test_remove_dataset():
    fed_client = federated.FederatedClient()

    dataset = "my_dataset"
    key = "string_dataset"
    fed_client.add_dataset(dataset, key)

    assert key in fed_client.datasets

    fed_client.remove_dataset(key)

    assert key not in fed_client.datasets


def test_set_obj_train_config():
    fed_client = federated.FederatedClient()
    fed_client.optimizer = True

    train_config = federated.TrainConfig(id=100, model=None, loss_fn=None)

    fed_client.set_obj(train_config)

    assert fed_client.train_config.id == train_config.id
    assert fed_client.optimizer is None


def test_set_obj_other():
    fed_client = federated.FederatedClient()

    dummy_data = torch.tensor(42)
    dummy_data.id = 43

    fed_client.set_obj(dummy_data)

    assert len(fed_client._objects) == 1
    assert fed_client._objects[dummy_data.id] == dummy_data


def evaluate_model(fed_client, model_id, loss_fn, data, target):  # pragma: no cover
    new_model = fed_client.get_obj(model_id)
    pred = new_model.obj(data)
    loss_after = loss_fn(target=target, pred=pred)
    return loss_after


def train_model(fed_client, fit_dataset_key, available_dataset_key, nr_rounds):  # pragma: no cover
    loss = None
    for curr_round in range(nr_rounds):
        if fit_dataset_key == available_dataset_key:
            loss = fed_client.fit(dataset_key=fit_dataset_key)
        else:
            with pytest.raises(ValueError):
                loss = fed_client.fit(dataset_key=fit_dataset_key)
        if PRINT_IN_UNITTESTS and curr_round % 2 == 0:  # pragma: no cover
            print("-" * 50)
            print("Iteration %s: alice's loss: %s" % (curr_round, loss))


@pytest.mark.parametrize(
    "fit_dataset_key, epochs",
    [("gaussian_mixture", 1), ("gaussian_mixture", 10), ("another_dataset", 1)],
)
def test_fit(fit_dataset_key, epochs):
    data, target = utils.create_gaussian_mixture_toy_data(nr_samples=100)

    fed_client = federated.FederatedClient()
    dataset = sy.BaseDataset(data, target)
    dataset_key = "gaussian_mixture"
    fed_client.add_dataset(dataset, key=dataset_key)

    def loss_fn(pred, target):
        return torch.nn.functional.cross_entropy(input=pred, target=target)

    class Net(torch.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = torch.nn.Linear(2, 3)
            self.fc2 = torch.nn.Linear(3, 2)

            torch.nn.init.xavier_normal_(self.fc1.weight)
            torch.nn.init.xavier_normal_(self.fc2.weight)

        def forward(self, x):
            x = torch.nn.functional.relu(self.fc1(x))
            x = torch.nn.functional.relu(self.fc2(x))
            return x

    model_untraced = Net()
    model = torch.jit.trace(model_untraced, data)
    model_id = 0
    model_ow = pointers.ObjectWrapper(obj=model, id=model_id)
    loss_id = 1
    loss_ow = pointers.ObjectWrapper(obj=loss_fn, id=loss_id)
    pred = model(data)
    loss_before = loss_fn(target=target, pred=pred)
    if PRINT_IN_UNITTESTS:  # pragma: no cover
        print("Loss before training: {}".format(loss_before))

    # Create and send train config
    train_config = sy.TrainConfig(
        batch_size=8,
        model=None,
        loss_fn=None,
        model_id=model_id,
        loss_fn_id=loss_id,
        optimizer_args={"lr": 0.05, "weight_decay": 0.01},
        epochs=epochs,
    )

    fed_client.set_obj(model_ow)
    fed_client.set_obj(loss_ow)
    fed_client.set_obj(train_config)
    fed_client.optimizer = None

    train_model(fed_client, fit_dataset_key, available_dataset_key=dataset_key, nr_rounds=3)

    if dataset_key == fit_dataset_key:
        loss_after = evaluate_model(fed_client, model_id, loss_fn, data, target)
        if PRINT_IN_UNITTESTS:  # pragma: no cover
            print("Loss after training: {}".format(loss_after))

        if loss_after >= loss_before:  # pragma: no cover
            if PRINT_IN_UNITTESTS:
                print("Loss not reduced, train more: {}".format(loss_after))

            train_model(
                fed_client, fit_dataset_key, available_dataset_key=dataset_key, nr_rounds=10
            )
            loss_after = evaluate_model(fed_client, model_id, loss_fn, data, target)

        assert loss_after < loss_before


@pytest.mark.skipif(
    torch.__version__ >= "1.1",
    reason="bug in pytorch version 1.1.0, jit.trace returns raw C function",
)
def test_evaluate():  # pragma: no cover
    data, target = utils.iris_data_partial()

    fed_client = federated.FederatedClient()
    dataset = sy.BaseDataset(data, target)
    dataset_key = "iris"
    fed_client.add_dataset(dataset, key=dataset_key)

    def loss_fn(pred, target):
        return torch.nn.functional.cross_entropy(input=pred, target=target)

    class Net(torch.nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.fc1 = torch.nn.Linear(4, 3)

        def forward(self, x):
            x = torch.nn.functional.relu(self.fc1(x))
            return x

    model_untraced = Net()

    with torch.no_grad():
        model_untraced.fc1.weight.set_(
            torch.tensor(
                [
                    [0.0160, 1.3753, -0.1202, -0.9129],
                    [0.1539, 0.3092, 0.0749, 0.2142],
                    [0.0984, 0.6248, 0.0274, 0.1735],
                ]
            )
        )
        model_untraced.fc1.bias.set_(torch.tensor([0.3477, 0.2970, -0.0799]))

    model = torch.jit.trace(model_untraced, data)
    model_id = 0
    model_ow = pointers.ObjectWrapper(obj=model, id=model_id)
    loss_id = 1
    loss_ow = pointers.ObjectWrapper(obj=loss_fn, id=loss_id)
    pred = model(data)
    loss_before = loss_fn(target=target, pred=pred)
    if PRINT_IN_UNITTESTS:  # pragma: no cover
        print("Loss before training: {}".format(loss_before))

    # Create and send train config
    train_config = sy.TrainConfig(
        batch_size=8,
        model=None,
        loss_fn=None,
        model_id=model_id,
        loss_fn_id=loss_id,
        optimizer_args=None,
        epochs=1,
    )

    fed_client.set_obj(model_ow)
    fed_client.set_obj(loss_ow)
    fed_client.set_obj(train_config)
    fed_client.optimizer = None

    result = fed_client.evaluate(
        dataset_key=dataset_key, return_histograms=True, nr_bins=3, return_loss=True
    )

    test_loss_before = result["loss"]
    correct_before = result["nr_correct_predictions"]
    len_dataset = result["nr_predictions"]
    hist_pred_before = result["histogram_predictions"]
    hist_target = result["histogram_target"]

    if PRINT_IN_UNITTESTS:  # pragma: no cover
        print("Evaluation result before training: {}".format(result))

    assert len_dataset == 30
    assert (hist_target == [10, 10, 10]).all()

    train_config = sy.TrainConfig(
        batch_size=8,
        model=None,
        loss_fn=None,
        model_id=model_id,
        loss_fn_id=loss_id,
        optimizer="SGD",
        optimizer_args={"lr": 0.01},
        shuffle=True,
        epochs=2,
    )
    fed_client.set_obj(train_config)
    train_model(fed_client, dataset_key, available_dataset_key=dataset_key, nr_rounds=50)

    result = fed_client.evaluate(
        dataset_key=dataset_key, return_histograms=True, nr_bins=3, return_loss=True
    )

    test_loss_after = result["loss"]
    correct_after = result["nr_correct_predictions"]
    len_dataset = result["nr_predictions"]
    hist_pred_after = result["histogram_predictions"]
    hist_target = result["histogram_target"]

    if PRINT_IN_UNITTESTS:  # pragma: no cover
        print("Evaluation result: {}".format(result))

    assert len_dataset == 30
    assert (hist_target == [10, 10, 10]).all()
    assert correct_after > correct_before
    assert torch.norm(torch.tensor(hist_target - hist_pred_after)) < torch.norm(
        torch.tensor(hist_target - hist_pred_before)
    )



import pytest
import torch

from syft.frameworks.torch.tensors.interpreters import LargePrecisionTensor


def test_wrap(workers):
    """
    Test the .on() wrap functionality for LargePrecisionTensor
    """
    x_tensor = torch.Tensor([1, 2, 3])
    x = LargePrecisionTensor().on(x_tensor)
    assert isinstance(x, torch.Tensor)
    assert isinstance(x.child, LargePrecisionTensor)
    assert isinstance(x.child.child, torch.Tensor)


def test_fix_prec(workers):
    x = torch.tensor([1.5, 2.0, 3.0])
    enlarged = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(x, restored))


def test_2d_tensors(workers):
    x = torch.tensor([[1.5, 2.0, 3.0], [4.5, 5.0, 6.0]])
    enlarged = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(x, restored))


def test_3d_tensors(workers):
    x = torch.tensor([[[1.5, 2.0, 3.0]], [[4.5, 5.0, 6.0]], [[7.0, 8.0, 9.0]]])
    enlarged = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(x, restored))


def test_negative_numbers(workers):
    x = torch.tensor([[[-1.5, 2.0, 3.0]], [[4.5, 5.0, 6.0]], [[7.0, 8.0, 9.0]]])
    enlarged = x.fix_prec(
        base=10, internal_type=torch.int16, precision_fractional=128, verbose=True
    )
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(x, restored))


def test_add_multiple_dimensions(workers):
    x = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.5, 5.0, -3.0]]])
    y = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.5, 5.0, 6.0]]])
    lpt1 = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    lpt2 = y.fix_prec(internal_type=torch.int16, precision_fractional=128)
    expected = torch.tensor([[[-3.0, -4.0, -6.0]], [[9.0, 10.0, 3.0]]])
    result = lpt1 + lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_add_negative_values():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([-90000000000000010.0])
    expected = torch.tensor([-90000000000000000.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 + lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_add():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([20.0])
    expected = torch.tensor([30.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 + lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_iadd():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([20.0])
    expected = torch.tensor([30.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1.add_(lpt2)
    assert torch.all(torch.eq(expected, lpt1.float_precision()))


def test_add_different_dims():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([100000.0])
    x2 = torch.tensor([20.0])
    expected = torch.tensor([100020.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 + lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_mul():
    internal_type = torch.int16
    precision_fractional = 32
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([20.0])
    expected = torch.tensor([200.0])
    expected.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 * lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_imul():
    internal_type = torch.int16
    precision_fractional = 32
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([20.0])
    expected = torch.tensor([200.0])
    expected.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1.mul_(lpt2)
    assert torch.all(torch.eq(expected, lpt1.float_precision()))


def test_mul_multiple_dims():
    internal_type = torch.int16
    precision_fractional = 32
    x = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.0, 5.0, -3.0]]])
    y = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.0, 5.0, 6.0]]])
    expected = torch.tensor([[[2.25, 4.0, 9.0]], [[16.0, 25.0, -18.0]]])
    expected.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1 = x.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = y.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 * lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_concat_ops():
    internal_type = torch.int16
    precision_fractional = 32
    x = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.0, 5.0, -3.0]]])
    y = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.0, 5.0, 6.0]]])
    z = torch.tensor([[[-1.0, -2.0, 2.5]], [[4.0, -5.0, 7.5]]])
    expected = torch.tensor([[[1.25, 2.0, 11.5]], [[20.0, 20.0, -10.5]]])
    expected.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt_x = x.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt_y = y.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt_z = z.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = (lpt_x * lpt_y) + lpt_z

    assert torch.all(torch.eq(expected, result.float_precision()))


def test_uint8_representation(workers):
    x = torch.tensor([[1.5, 2.0, 3.0], [4.5, 5.0, 6.0]])
    enlarged = x.fix_prec(internal_type=torch.uint8, precision_fractional=128)
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(x, restored))


def test_sub():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([90000000000000010.0])
    expected = torch.tensor([-90000000000000000.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 - lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_isub():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([90000000000000010.0])
    expected = torch.tensor([-90000000000000000.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1.sub_(lpt2)
    assert torch.all(torch.eq(expected, lpt1.float_precision()))


def test_diff_dims_in_same_tensor():
    internal_type = torch.int16
    precision_fractional = 128
    x = torch.tensor([2000.0, 1.0])
    lpt_x = x.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    restored = lpt_x.float_precision()
    assert torch.all(torch.eq(x, restored))


def test_mod():
    internal_type = torch.int16
    precision_fractional = 128
    expected = torch.tensor([6.0, 3.0])
    x1 = torch.tensor([6.0, 12.0])
    x2 = torch.tensor([9.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 % lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


@pytest.mark.parametrize(
    "x, expected",
    [
        (torch.tensor([1]), torch.tensor([1.0])),
        (torch.tensor([1.0]), torch.tensor([1.0])),
        (torch.tensor([2000.0, 1.0]), torch.tensor([2000.0, 1.0])),
        (torch.tensor([2000.0, 1]), torch.tensor([2000.0, 1.0])),
        (torch.tensor([-2000.0]), torch.tensor([-2000.0])),
        (torch.tensor([-2000.0, -50]), torch.tensor([-2000.0, -50.0])),
        (torch.tensor([-2000.0, 50]), torch.tensor([-2000.0, 50.0])),
        (
            torch.tensor([[-2000.0, 50], [1000.5, -25]]),
            torch.tensor([[-2000.0, 50.0], [1000.5, -25.0]]),
        ),
        (torch.tensor([-2000.0123458910]), torch.tensor([-2000.0123458910])),
        (torch.tensor([2000.0123458910]), torch.tensor([2000.0123458910])),
    ],
)
def test_types(x, expected):
    enlarged = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(expected, restored))



from collections import defaultdict, OrderedDict
import inspect

import tensorflow as tf
import tf_encrypted as tfe

# When instantiating tfe layers, exclude not supported arguments in TFE.
_args_not_supported_by_tfe = [
    "activity_regularizer",
    "kernel_regularizer",
    "bias_regularizer",
    "kernel_constraint",
    "bias_constraint",
    "dilation_rate",
]


def share(model, cluster, target_graph=None):
    """
    Secret share the model between `workers`.

    This is done by rebuilding the model as a TF Encrypted model inside
    `target_graph` and pushing this graph to TFEWorkers running the cluster.

    Note that this keeps a TensorFlow session alive that must later be
    shutdown using `model.stop()`.
    """

    # Store Keras weights before loading them in the TFE layers.
    # TODO(Morten) we could optimize runtime by running a single model.get_weights instead
    stored_keras_weights = {
        keras_layer.name: keras_layer.get_weights() for keras_layer in model.layers
    }

    # Handle input combinations to configure TFE
    _configure_tfe(cluster)

    if target_graph is None:
        # By default we create a new graph for the shared model
        target_graph = tf.Graph()

    model._tfe_graph = target_graph
    with model._tfe_graph.as_default():
        tfe_model, batch_input_shape = _rebuild_tfe_model(model, stored_keras_weights)

        # Set up a new tfe.serving.QueueServer for the shared TFE model
        q_input_shape = batch_input_shape
        q_output_shape = model.output_shape

        model._queue_server = tfe.serving.QueueServer(
            input_shape=q_input_shape, output_shape=q_output_shape, computation_fn=tfe_model
        )

        initializer = tf.global_variables_initializer()

    # Push and initialize shared model on servers
    model._tfe_session = tfe.Session(graph=model._tfe_graph)
    model._tfe_session.run(initializer)


def serve(model, num_requests=5):
    """
    Serve the specified number of predictions using the shared model,
    blocking until completed.
    """

    global request_ix
    request_ix = 1

    def step_fn():
        global request_ix
        print("Served encrypted prediction {i} to client.".format(i=request_ix))
        request_ix += 1

    model._queue_server.run(model._tfe_session, num_steps=num_requests, step_fn=step_fn)


def stop(model):
    """
    Shutdown the TensorFlow session that was used to serve the shared model.
    """

    if model._tfe_session is not None:
        sess = model._tfe_session
        model._tfe_session = None
        sess.close()
        del sess


def _configure_tfe(cluster):

    if not cluster or len(cluster.workers) != 3:
        raise RuntimeError("TF Encrypted expects three parties for its sharing protocols.")

    config = cluster.tfe_config
    tfe.set_config(config)

    prot = tfe.protocol.SecureNN(
        config.get_player("server0"), config.get_player("server1"), config.get_player("server2")
    )
    tfe.set_protocol(prot)
    tfe.clear_initializers()


def _rebuild_tfe_model(keras_model, stored_keras_weights):
    """
    Rebuild the plaintext Keras model as a TF Encrypted Keras model
    from the plaintext weights in `stored_keras_weights` using the
    current TensorFlow graph, and the current TF Encrypted protocol
    and configuration.
    """

    tfe_model = tfe.keras.Sequential()

    for keras_layer in keras_model.layers:
        tfe_layer = _instantiate_tfe_layer(keras_layer, stored_keras_weights)
        tfe_model.add(tfe_layer)

        if hasattr(tfe_layer, "_batch_input_shape"):
            batch_input_shape = tfe_layer._batch_input_shape

    return tfe_model, batch_input_shape


def _instantiate_tfe_layer(keras_layer, stored_keras_weights):

    # Get original layer's constructor parameters
    # This method is added by the KerasHook object in syft.keras
    constructor_params = keras_layer._constructor_parameters_store
    constructor_params.apply_defaults()
    _trim_params(constructor_params.arguments, _args_not_supported_by_tfe + ["self"])

    # Identify tf.keras layer type, and grab the corresponding tfe.keras layer
    keras_layer_type = _get_layer_type(keras_layer)
    try:
        tfe_layer_cls = getattr(tfe.keras.layers, keras_layer_type)
    except AttributeError:
        # TODO: rethink how we warn the user about this, maybe codegen a list of
        #       supported layers in a doc somewhere
        raise RuntimeError(
            "TF Encrypted does not yet support the " "{lcls} layer.".format(lcls=keras_layer_type)
        )

    # Extract argument list expected by layer __init__
    # TODO[jason]: find a better way
    tfe_params = list(inspect.signature(tfe_layer_cls.__init__).parameters.keys())

    # Remove arguments currently not supported by TFE layers
    _trim_params(tfe_params, _args_not_supported_by_tfe + ["self", "kwargs"])

    # Load weights from Keras layer into TFE layer
    # TODO[jason]: generalize to n weights -- will require special handling for
    #              optional weights like bias (or batchnorm trainables)
    if "kernel_initializer" in constructor_params.arguments:
        kernel_weights = stored_keras_weights[keras_layer.name][0]
        k_initializer = tf.keras.initializers.Constant(kernel_weights)
        constructor_params.arguments["kernel_initializer"] = k_initializer

    if "use_bias" in constructor_params.arguments:
        if constructor_params.arguments["use_bias"]:
            bias_weights = stored_keras_weights[keras_layer.name][1]
            b_initializer = tf.keras.initializers.Constant(bias_weights)
            constructor_params.arguments["bias_initializer"] = b_initializer

    unpacked_kwargs = constructor_params.arguments.pop("kwargs")
    tfe_kwargs = {**constructor_params.arguments, **unpacked_kwargs}

    return tfe_layer_cls(**tfe_kwargs)


def _get_layer_type(keras_layer_cls):
    return keras_layer_cls.__class__.__name__


def _trim_params(params, filter_list):
    for arg_name in filter_list:
        try:
            del params[arg_name]
        except TypeError:
            # params is a list of strings
            if arg_name in params:
                params.remove(arg_name)
        except KeyError:
            continue



import random
from typing import List
from syft import exceptions


def create_random_id():
    return int(10e10 * random.random())


class IdProvider:
    """Provides Id to all syft objects.

    Generate id and store the list of ids generated
    Can take a pre set list in input and will complete
    when it's empty.

    An instance of IdProvider is accessible via sy.ID_PROVIDER.
    """

    def __init__(self, given_ids=None):
        self.given_ids = given_ids if given_ids is not None else list()
        self.generated = set()
        self.record_ids = False
        self.recorded_ids = []

    def pop(self, *args) -> int:
        """Provides random ids and store them.

        The syntax .pop() mimics the list syntax for convenience
        and not the generator syntax.

        Returns:
            Random Id.
        """
        if len(self.given_ids):
            random_id = self.given_ids.pop(-1)
        else:
            random_id = create_random_id()
            while random_id in self.generated:
                random_id = create_random_id()
        self.generated.add(random_id)
        if self.record_ids:
            self.recorded_ids.append(random_id)

        return random_id

    def set_next_ids(self, given_ids: List, check_ids: bool = True):
        """Sets the next ids returned by the id provider

        Note that the ids are returned in reverse order of the list, as a pop() operation is applied.

        Args:
            given_ids: List, next ids returned by the id provider
            check_ids: bool, check whether these ids conflict with already generated ids

        """
        if check_ids:
            intersect = self.generated.intersection(set(given_ids))
            if len(intersect) > 0:
                message = "Provided IDs {} are contained in already generated IDs".format(intersect)
                raise exceptions.IdNotUniqueError(message)

        self.given_ids += given_ids

    def start_recording_ids(self):
        """Starts the recording in form of a list of the generated ids.
        """
        self.record_ids = True
        self.recorded_ids = list()

    def get_recorded_ids(self, continue_recording=False):
        """Returns the generated ids since the last call to start_recording_ids.

        Args:
            continue_recording: if False, the recording is stopped and the list of recorded ids is reset

        Returns:
            list of recorded ids
        """
        ret_val = self.recorded_ids
        if not continue_recording:
            self.record_ids = False
            self.recorded_ids = list()
        return ret_val



import numpy as np

import torch

from syft.frameworks.torch.differential_privacy import pate

np.random.seed(0)


def test_base_dataset():

    num_teachers, num_examples, num_labels = (100, 50, 10)
    preds = (np.random.rand(num_teachers, num_examples) * num_labels).astype(int)  # fake preds

    indices = (np.random.rand(num_examples) * num_labels).astype(int)  # true answers

    preds[:, 0:10] *= 0

    data_dep_eps, data_ind_eps = pate.perform_analysis(
        teacher_preds=preds, indices=indices, noise_eps=0.1, delta=1e-5
    )

    assert data_dep_eps < data_ind_eps


def test_base_dataset_torch():

    num_teachers, num_examples, num_labels = (100, 50, 10)
    preds = (np.random.rand(num_teachers, num_examples) * num_labels).astype(int)  # fake preds

    indices = (np.random.rand(num_examples) * num_labels).astype(int)  # true answers

    preds[:, 0:10] *= 0

    data_dep_eps, data_ind_eps = pate.perform_analysis_torch(
        preds, indices, noise_eps=0.1, delta=1e-5
    )

    assert data_dep_eps < data_ind_eps


def test_torch_ref_match():

    # Verify if the torch implementation values match the original Numpy implementation.

    num_teachers, num_examples, num_labels = (100, 50, 10)
    preds = (np.random.rand(num_teachers, num_examples) * num_labels).astype(int)  # fake preds

    indices = (np.random.rand(num_examples) * num_labels).astype(int)  # true answers

    preds[:, 0:10] *= 0

    data_dep_eps, data_ind_eps = pate.perform_analysis_torch(
        preds, indices, noise_eps=0.1, delta=1e-5
    )

    data_dep_eps_ref, data_ind_eps_ref = pate.perform_analysis(
        preds, indices, noise_eps=0.1, delta=1e-5
    )

    assert torch.isclose(data_dep_eps, torch.tensor(data_dep_eps_ref.item()))
    assert torch.isclose(data_ind_eps, torch.tensor(data_ind_eps_ref.item()))



from syft.federated.train_config import TrainConfig
from syft.federated import federated_client
from syft.federated.federated_client import FederatedClient

__all__ = ["TrainConfig", "federated_client"]



from time import time
from unittest.mock import patch

import syft as sy
from syft.exceptions import GetNotPermittedError
from syft.workers.virtual import VirtualWorker
from syft.codes import MSGTYPE
from syft import messaging
from syft import serde

from syft.frameworks.torch import pointers

import pytest
import torch
import torch as th


def test_send_msg():
    """Tests sending a message with a specific ID

    This is a simple test to ensure that the BaseWorker interface
    can properly send/receive a message containing a tensor.
    """

    # get pointer to local worker
    me = sy.torch.hook.local_worker

    # create a new worker (to send the object to)
    worker_id = sy.ID_PROVIDER.pop()
    bob = VirtualWorker(sy.torch.hook, id=f"bob{worker_id}")

    # initialize the object and save it's id
    obj = torch.Tensor([100, 100])
    obj_id = obj.id

    # Send data to bob
    me.send_msg(MSGTYPE.OBJ, obj, bob)

    # ensure that object is now on bob's machine
    assert obj_id in bob._objects


def test_send_msg_using_tensor_api():
    """Tests sending a message with a specific ID

    This is a simple test to ensure that the high level tensor .send()
    method correctly sends a message to another worker.
    """

    # create worker to send object to
    worker_id = sy.ID_PROVIDER.pop()
    bob = VirtualWorker(sy.torch.hook, id=f"bob{worker_id}")

    # create a tensor to send (default on local_worker)
    obj = torch.Tensor([100, 100])

    # save the object's id
    obj_id = obj.id

    # send the object to Bob (from local_worker)
    _ = obj.send(bob)

    # ensure tensor made it to Bob
    assert obj_id in bob._objects


def test_recv_msg():
    """Tests the recv_msg command with 2 tests

    The first test uses recv_msg to send an object to alice.

    The second test uses recv_msg to request the object
    previously sent to alice."""

    # TEST 1: send tensor to alice

    # create a worker to send data to
    worker_id = sy.ID_PROVIDER.pop()
    alice = VirtualWorker(sy.torch.hook, id=f"alice{worker_id}")

    # create object to send
    obj = torch.Tensor([100, 100])

    # create/serialize message
    message = messaging.Message(MSGTYPE.OBJ, obj)
    bin_msg = serde.serialize(message)

    # have alice receive message
    alice.recv_msg(bin_msg)

    # ensure that object is now in alice's registry
    assert obj.id in alice._objects

    # Test 2: get tensor back from alice

    # Create message: Get tensor from alice
    message = messaging.Message(MSGTYPE.OBJ_REQ, obj.id)

    # serialize message
    bin_msg = serde.serialize(message)

    # call receive message on alice
    resp = alice.recv_msg(bin_msg)

    obj_2 = serde.deserialize(resp)

    # assert that response is correct type
    assert type(resp) == bytes

    # ensure that the object we receive is correct
    assert obj_2.id == obj.id


def tests_worker_convenience_methods():
    """Tests send and get object methods on BaseWorker

    This test comes in two parts. The first uses the simple
    BaseWorker.send_obj and BaseWorker.request_obj to send a
    tensor to Alice and to get the worker back from Alice.

    The second part shows that the same methods work between
    bob and alice directly.
    """

    me = sy.torch.hook.local_worker
    worker_id = sy.ID_PROVIDER.pop()
    bob = VirtualWorker(sy.torch.hook, id=f"bob{worker_id}")
    worker_id = sy.ID_PROVIDER.pop()
    alice = VirtualWorker(sy.torch.hook, id=f"alice{worker_id}")
    obj = torch.Tensor([100, 100])

    # Send data to alice
    me.send_obj(obj, alice)

    # Get data from alice
    resp_alice = me.request_obj(obj.id, alice)

    assert (resp_alice == obj).all()

    obj2 = torch.Tensor([200, 200])

    # Set data on self
    bob.set_obj(obj2)

    # Get data from self
    resp_bob_self = bob.get_obj(obj2.id)

    assert (resp_bob_self == obj2).all()

    # Get data from bob as alice
    resp_bob_alice = alice.request_obj(obj2.id, bob)

    assert (resp_bob_alice == obj2).all()


def test_search():
    worker_id = sy.ID_PROVIDER.pop()
    bob = VirtualWorker(sy.torch.hook, id=f"bob{worker_id}")

    x = (
        torch.tensor([1, 2, 3, 4, 5])
        .tag("#fun", "#mnist")
        .describe("The images in the MNIST training dataset.")
        .send(bob)
    )

    y = (
        torch.tensor([1, 2, 3, 4, 5])
        .tag("#not_fun", "#cifar")
        .describe("The images in the MNIST training dataset.")
        .send(bob)
    )

    z = (
        torch.tensor([1, 2, 3, 4, 5])
        .tag("#fun", "#boston_housing")
        .describe("The images in the MNIST training dataset.")
        .send(bob)
    )

    a = (
        torch.tensor([1, 2, 3, 4, 5])
        .tag("#not_fun", "#boston_housing")
        .describe("The images in the MNIST training dataset.")
        .send(bob)
    )

    assert len(bob.search("#fun")) == 2
    assert len(bob.search("#mnist")) == 1
    assert len(bob.search("#cifar")) == 1
    assert len(bob.search("#not_fun")) == 2
    assert len(bob.search("#not_fun", "#boston_housing")) == 1


def test_obj_not_found(workers):
    """Test for useful error message when trying to call a method on
    a tensor which does not exist on a worker anymore."""

    bob = workers["bob"]

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)

    bob._objects = {}

    try:
        y = x + x
    except KeyError as e:
        assert "If you think this tensor does exist" in str(e)


def test_get_not_permitted(workers):
    bob = workers["bob"]
    with patch.object(th.Tensor, "allowed_to_get") as mock_allowed_to_get:
        mock_allowed_to_get.return_value = False
        x = th.tensor([1, 2, 3, 4, 5]).send(bob)
        with pytest.raises(GetNotPermittedError):
            x.get()
        mock_allowed_to_get.assert_called_once()


def test_spinup_time(hook):
    """Tests to ensure that virtual workers intialized with 10000 data points
    load in under 0.05 seconds. This is needed to ensure that virtual workers
    spun up inside web frameworks are created quickly enough to not cause timeout errors"""
    data = []
    for i in range(10000):
        data.append(th.Tensor(5, 5).random_(100))
    start_time = time()
    dummy = sy.VirtualWorker(hook, id="dummy", data=data)
    end_time = time()
    assert (end_time - start_time) < 0.05


@pytest.mark.skipif(
    torch.__version__ >= "1.1",
    reason="bug in pytorch version 1.1.0, jit.trace returns raw C function",
)
def test_send_jit_scriptmodule(hook, workers):  # pragma: no cover
    bob = workers["bob"]

    @torch.jit.script
    def foo(x):
        return x + 2

    foo_wrapper = pointers.ObjectWrapper(obj=foo, id=99)
    foo_ptr = hook.local_worker.send(foo_wrapper, bob)

    res = foo_ptr(torch.tensor(4))
    assert res == torch.tensor(6)



import sys
import pytest
from syft import dependency_check


@pytest.mark.skipif(not dependency_check.keras_available, reason="tf_encrypted not installed")
def test_tf_encrypted_available():
    sys.modules.pop("syft", None)
    sys.modules.pop("syft.dependency_check", None)
    from syft import dependency_check

    assert dependency_check.keras_available


@pytest.mark.usefixtures("no_tf_encrypted")
def test_tf_encrypted_missing():

    sys.modules.pop("syft.dependency_check", None)
    sys.modules.pop("tf_encrypted", None)
    sys.modules.pop("tfe", None)
    import syft.dependency_check

    assert not syft.dependency_check.keras_available



from syft.frameworks.keras.model.sequential import serve
from syft.frameworks.keras.model.sequential import share
from syft.frameworks.keras.model.sequential import stop

__all__ = ["serve", "share", "stop"]



import logging
from syft import dependency_check

logger = logging.getLogger(__name__)

__all__ = list()

if dependency_check.keras_available:
    from . import keras

    __all__.append("keras")

from . import torch

__all__.append("torch")



import syft as sy

from syft.workers import AbstractWorker


class Message:
    def __init__(self, msg_type, contents):
        self.msg_type = msg_type
        self.contents = contents

    def _simplify(self):
        return (self.msg_type, self.contents)

    @staticmethod
    def simplify(ptr: "Message") -> tuple:
        """
        This function takes the attributes of a Message and saves them in a tuple
        Args:
            ptr (Message): a Message
        Returns:
            tuple: a tuple holding the unique attributes of the message
        Examples:
            data = simplify(ptr)
        """

        return (ptr.msg_type, sy.serde._simplify(ptr.contents))

    @staticmethod
    def detail(worker: AbstractWorker, tensor_tuple: tuple) -> "Message":
        return Message(tensor_tuple[0], sy.serde._detail(worker, tensor_tuple[1]))



import syft as sy


def test_message_serde(hook):

    x = sy.Message(0, [1, 2, 3])
    x_bin = sy.serde.serialize(x)
    y = sy.serde.deserialize(x_bin, sy.local_worker)

    assert x.contents == y.contents
    assert x.msg_type == y.msg_type



import pytest
import torch
from multiprocessing import Process
import builtins

import syft
from syft import TorchHook


@pytest.fixture()
def start_proc():  # pragma: no cover
    """ helper function for spinning up a websocket participant """

    def _start_proc(participant, dataset: str = None, **kwargs):
        def target():
            server = participant(**kwargs)
            if dataset is not None:
                data, key = dataset
                server.add_dataset(data, key=key)
            server.start()

        p = Process(target=target)
        p.start()
        return p

    return _start_proc


@pytest.fixture(scope="session", autouse=True)
def hook():
    hook = TorchHook(torch)
    return hook


@pytest.fixture(scope="function", autouse=True)
def workers(hook):
    # To run a plan locally the local worker can't be a client worker,
    # since it needs to register objects
    # LaRiffle edit: doing this increases the reference count on pointers and
    # breaks the auto garbage collection for pointer of pointers, see #2150
    # hook.local_worker.is_client_worker = False

    # Reset the hook and the local worker
    syft.local_worker.clear_objects()
    syft.frameworks.torch.hook.hook_args.hook_method_args_functions = {}
    syft.frameworks.torch.hook.hook_args.hook_method_response_functions = {}
    syft.frameworks.torch.hook.hook_args.get_tensor_type_functions = {}

    # Define 3 virtual workers
    alice = syft.VirtualWorker(id="alice", hook=hook, is_client_worker=False)
    bob = syft.VirtualWorker(id="bob", hook=hook, is_client_worker=False)
    james = syft.VirtualWorker(id="james", hook=hook, is_client_worker=False)

    workers = {"me": hook.local_worker, "alice": alice, "bob": bob, "james": james}

    yield workers

    alice.remove_worker_from_local_worker_registry()
    bob.remove_worker_from_local_worker_registry()
    james.remove_worker_from_local_worker_registry()


@pytest.fixture
def no_tf_encrypted():
    import_orig = builtins.__import__

    def mocked_import(name, globals, locals, fromlist, level):
        if "tf_encrypted" in name:
            raise ImportError()
        return import_orig(name, globals, locals, fromlist, level)

    builtins.__import__ = mocked_import
    yield
    builtins.__import__ = import_orig



import os

from setuptools import find_packages
from setuptools import setup


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


requirements = read("requirements.txt").split()


setup(
    name="syft",
    version="0.1.22a1",
    author="Andrew Trask",
    author_email="contact@openmined.org",
    description=("A Library for Private, Secure Deep Learning"),
    license="Apache-2.0",
    keywords="deep learning artificial intelligence privacy secure multi-party computation federated learning differential privacy",
    packages=find_packages(exclude=["docs", "examples", "dist"]),
    include_package_data=True,
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/OpenMined/PySyft",
    install_requires=requirements,
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "pytest-flake8"],
    classifiers=["Programming Language :: Python :: 3", "Operating System :: OS Independent"],
)



import pytest
import torch
import torch.nn.functional as F

from syft.frameworks.torch.tensors.decorators import LoggingTensor


def test_wrap():
    """
    Test the .on() wrap functionality for LoggingTensor
    """

    x_tensor = torch.Tensor([1, 2, 3])
    x = LoggingTensor().on(x_tensor)

    assert isinstance(x, torch.Tensor)
    assert isinstance(x.child, LoggingTensor)
    assert isinstance(x.child.child, torch.Tensor)


def test_overwritten_method_on_log_chain():
    """
    Test method call on a chain including a log tensor
    """

    # Build a long chain tensor Wrapper>LoggingTensor>TorchTensor
    x_tensor = torch.Tensor([1, 2, 3])
    x = LoggingTensor().on(x_tensor)
    y = x.add(x)

    assert (y.child.child == x_tensor.add(x_tensor)).all()

    y = x.child.manual_add(x.child)

    assert (y.child == x_tensor.add(x_tensor)).all()


def test_method_on_log_chain():
    """
    Test method call on a chain including a log tensor
    """

    # Build a long chain tensor Wrapper>LoggingTensor>TorchTensor
    x_tensor = torch.Tensor([1, 2, 3])
    x = LoggingTensor().on(x_tensor)
    y = x.mul(x)

    assert (y.child.child == x_tensor.mul(x_tensor)).all()


@pytest.mark.parametrize("attr", ["relu", "celu", "elu"])
def test_hook_module_functional_on_log_chain(attr):
    """
    Test torch function call on a chain including a log tensor
    """

    attr = getattr(F, attr)
    x = torch.Tensor([1, -1, 3, 4])
    expected = attr(x)

    x_log = LoggingTensor().on(x)
    res_log = attr(x_log)
    res = res_log.child.child

    assert (res == expected).all()


def test_function_on_log_chain():
    """
    Test torch function call on a chain including a log tensor
    """

    x = LoggingTensor().on(torch.Tensor([1, -1, 3]))
    y = F.relu(x)

    assert (y.child.child == torch.Tensor([1, 0, 3])).all()


def test_send_get_log_chain(workers):
    """
    Test sending and getting back a chain including a logtensor
    """

    # Build a long chain tensor Wrapper>LoggingTensor>TorchTensor
    x_tensor = torch.Tensor([1, 2, 3])
    x = LoggingTensor().on(x_tensor)
    x_ptr = x.send(workers["bob"])
    x_back = x_ptr.get()

    assert (x_back.child.child == x_tensor).all()


def test_inplace_send_get_log_chain(workers):
    """
    Test sending and getting back a chain including a logtensor
    """

    # Build a long chain tensor Wrapper>LoggingTensor>TorchTensor
    x_tensor = torch.Tensor([1, 2, 3])
    x = LoggingTensor().on(x_tensor)
    x_ptr = x.send_(workers["bob"])
    x_back = x_ptr.get_()

    assert (x_back.child.child == x_tensor).all()


def test_remote_method_on_log_chain(workers):
    """
    Test remote method call on a chain including a log tensor
    """

    # Build a long chain tensor Wrapper>LoggingTensor>TorchTensor
    x_tensor = torch.Tensor([1, 2, 3])
    x = LoggingTensor().on(x_tensor)
    x_ptr = x.send(workers["bob"])
    y_ptr = F.relu(x_ptr)
    y = y_ptr.get()

    assert (y.child.child == F.relu(x_tensor)).all()


def test_remote_function_on_log_chain(workers):
    """
    Test remote function call on a chain including a log tensor
    """

    # Build a long chain tensor Wrapper>LoggingTensor>TorchTensor
    x_tensor = torch.Tensor([1, 2, 3])
    x = LoggingTensor().on(x_tensor)
    x_ptr = x.send(workers["bob"])
    y_ptr = x_ptr.add(x_ptr)
    y = y_ptr.get()

    assert (y.child.child == x_tensor.add(x_tensor)).all()


def test_print_log_chain():
    """
    Test sending and getting back a chain including a logtensor
    """

    # Build a long chain tensor Wrapper>LoggingTensor>TorchTensor
    x_tensor = torch.Tensor([1, 2, 3])
    x = LoggingTensor().on(x_tensor)

    assert isinstance(x.__str__(), str)
    assert isinstance(x.__repr__(), str)



import sys
import pytest


@pytest.mark.usefixtures("no_tf_encrypted")
def test_tf_encrypted_missing_import_syft():

    sys.modules.pop("syft", None)
    sys.modules.pop("syft.dependency_check", None)
    sys.modules.pop("tf_encrypted", None)
    sys.modules.pop("tfe", None)
    import syft

    with pytest.raises(AttributeError):
        kerasHook = syft.KerasHook

    with pytest.raises(AttributeError):
        tfeWorker = syft.TFEWorker



from collections import Counter
from typing import Tuple
from typing import Dict


class VirtualGrid:
    def __init__(self, *workers):
        self.workers = workers

    def search(
        self, *query, verbose: bool = True, return_counter: bool = True
    ) -> Tuple[Dict, Counter]:
        """Searches over a collection of workers, returning pointers to the results
        grouped by worker."""

        tag_counter = Counter()
        result_counter = 0

        results = {}
        for worker in self.workers:

            worker_tag_ctr = Counter()

            worker_results = worker.search(*query)

            if len(worker_results) > 0:
                results[worker.id] = worker_results

                for result in worker_results:
                    for tag in result.tags:
                        tag_counter[tag] += 1
                        worker_tag_ctr[tag] += 1

                if verbose:
                    tags = str(worker_tag_ctr.most_common(3))
                    print(
                        "Found "
                        + str(len(worker_results))
                        + " results on "
                        + str(worker)
                        + " - "
                        + tags
                    )

                result_counter += len(worker_results)

        if verbose:
            print("\nFound " + str(result_counter) + " results in total.")
            print("\nTag Profile:")
            for tag, count in tag_counter.most_common():
                print("\t" + tag + " found " + str(count))

        if return_counter:
            return results, tag_counter
        else:
            return results



import pytest
import torch
import syft as sy

# Import Hook
from syft.frameworks.torch import TorchHook

# Import grids
from syft.grid import VirtualGrid


def test_sandbox():
    sy.create_sandbox(globals(), download_data=False)

    assert alice == alice  # noqa: F821
    assert isinstance(alice, sy.VirtualWorker)  # noqa: F821
    assert andy == andy  # noqa: F821
    assert isinstance(andy, sy.VirtualWorker)  # noqa: F821
    assert bob == bob  # noqa: F821
    assert isinstance(bob, sy.VirtualWorker)  # noqa: F821
    assert jason == jason  # noqa: F821
    assert isinstance(jason, sy.VirtualWorker)  # noqa: F821
    assert jon == jon  # noqa: F821
    assert isinstance(jon, sy.VirtualWorker)  # noqa: F821
    assert theo == theo  # noqa: F821
    assert isinstance(theo, sy.VirtualWorker)  # noqa: F821

    assert hook == hook  # noqa: F821
    assert isinstance(hook, TorchHook)  # noqa: F821

    assert grid == grid  # noqa: F821
    assert isinstance(grid, VirtualGrid)  # noqa: F821

    assert workers == [bob, theo, jason, alice, andy, jon]  # noqa: F821



import pytest

import torch

from syft.frameworks.torch.pointers import PointerTensor
from syft.exceptions import InvalidTensorForRemoteGet


def test___str__(workers):
    bob = workers["bob"]
    tensor = torch.Tensor([1, 2, 3, 4])
    assert isinstance(tensor.__str__(), str)

    tensor_ptr = tensor.send(bob)
    assert isinstance(tensor_ptr.__str__(), str)


def test___repr__(workers):
    bob = workers["bob"]

    tensor = torch.Tensor([1, 2, 3, 4])
    assert isinstance(tensor.__repr__(), str)

    tensor_ptr = tensor.send(bob)
    assert isinstance(tensor_ptr.__repr__(), str)

    tensor = torch.Tensor([1, 2, 3, 4]).tag("#my_tag").describe("This is a description")
    assert isinstance(tensor.__repr__(), str)


def test_overload_reshape():
    tensor = torch.Tensor([1, 2, 3, 4])
    tensor_reshaped = tensor.reshape((2, 2))
    tensor_matrix = torch.Tensor([[1.0, 2.0], [3.0, 4.0]])
    assert (tensor_reshaped == tensor_matrix).all()


def test_owner_default(hook):
    tensor = torch.Tensor([1, 2, 3, 4, 5])
    assert tensor.owner == hook.local_worker


def test_create_pointer(hook, workers):
    bob = workers["bob"]

    tensor = torch.Tensor([1, 2, 3, 4, 5])

    ptr = tensor.create_pointer(
        location=bob, id_at_location=1, register=False, owner=hook.local_worker, ptr_id=2
    )

    assert ptr.owner == hook.local_worker
    assert ptr.location == bob
    assert ptr.id_at_location == 1
    assert ptr.id == 2

    ptr2 = tensor.create_pointer(owner=hook.local_worker)
    assert isinstance(ptr2.__str__(), str)
    assert isinstance(ptr2.__repr__(), str)


def test_create_pointer_defaults(workers):
    bob = workers["bob"]

    tensor = torch.Tensor([1, 2, 3, 4, 5])

    ptr = tensor.create_pointer(location=bob)

    assert ptr.owner == tensor.owner
    assert ptr.location == bob


def test_get(workers):
    bob = workers["bob"]

    tensor = torch.rand(5, 3)
    pointer = tensor.send(bob)

    assert type(pointer.child) == PointerTensor
    assert (pointer.get() == tensor).all()


def test_invalid_remote_get(workers):
    bob = workers["bob"]

    tensor = torch.rand(5, 3)
    pointer = tensor.send(bob)
    with pytest.raises(InvalidTensorForRemoteGet):
        pointer.remote_get()


def test_remote_get(hook, workers):
    me = workers["me"]
    bob = workers["bob"]
    alice = workers["alice"]

    x = torch.tensor([1, 2, 3, 4, 5])
    ptr_ptr_x = x.send(bob).send(alice)

    assert ptr_ptr_x.owner == me
    assert ptr_ptr_x.location == alice
    assert x.id in bob._objects

    assert len(bob._objects) == 1
    assert len(alice._objects) == 1

    ptr_ptr_x.remote_get()

    assert len(bob._objects) == 0
    assert len(alice._objects) == 1


def test_copy():
    tensor = torch.rand(5, 3)
    coppied_tensor = tensor.copy()
    assert (tensor == coppied_tensor).all()
    assert tensor is not coppied_tensor


def test_size():
    tensor = torch.rand(5, 3)
    assert tensor.size() == torch.Size([5, 3])
    assert tensor.size() == tensor.shape
    assert tensor.size(0) == tensor.shape[0]


# Compare local dim with the remote one
def test_dim(workers):
    tensor_local = torch.randn(5, 3)
    tensor_remote = tensor_local.send(workers["alice"])

    assert tensor_local.dim() == tensor_remote.dim()


def test_does_not_require_large_precision():
    x = torch.tensor([[[-1.5, 2.0, 30000000000.0]], [[4.5, 5.0, 6.0]], [[7.0, 8.0, 9.0]]])
    base = 10
    prec_fractional = 3
    max_precision = 62
    assert not x._requires_large_precision(max_precision, base, prec_fractional)


def test_requires_large_precision():
    x = torch.tensor([[[-1.5, 2.0, 30000000000.0]], [[4.5, 5.0, 6.0]], [[7.0, 8.0, 9.0]]])
    base = 10
    prec_fractional = 256
    max_precision = 62
    assert x._requires_large_precision(max_precision, base, prec_fractional)


def test_roll(workers):
    x = torch.tensor([1.0, 2.0, 3, 4, 5])
    expected = torch.roll(x, -1)

    index = torch.tensor([-1.0])
    result = torch.roll(x, index)

    assert (result == expected).all()



import pytest
import torch as th
import syft as sy

from syft.frameworks.torch.tensors.decorators import LoggingTensor


def test_tensors_not_collated_exception(workers):
    """
    Ensure that the sy.combine_pointers works as expected
    """

    bob = workers["bob"]
    alice = workers["alice"]

    x = th.tensor([1, 2, 3, 4, 5]).send(bob)
    y = th.tensor([1, 2, 3, 4, 5]).send(alice)

    with pytest.raises(sy.exceptions.TensorsNotCollocatedException):
        b = x + y

    x = th.tensor([1, 2, 3, 4, 5]).send(alice)
    y = th.tensor([1, 2, 3, 4, 5]).send(bob)

    with pytest.raises(sy.exceptions.TensorsNotCollocatedException):
        b = x + y

    x = th.tensor([1, 2, 3, 4, 5]).send(alice)
    y = th.tensor([1, 2, 3, 4, 5])

    with pytest.raises(sy.exceptions.TensorsNotCollocatedException):
        b = x + y

    x = th.tensor([1, 2, 3, 4, 5])
    y = th.tensor([1, 2, 3, 4, 5]).send(alice)

    with pytest.raises(sy.exceptions.TensorsNotCollocatedException):
        b = x + y



import pytest

import torch
import torch.nn as nn
import torch as th
import syft

from syft.frameworks.torch.tensors.interpreters import AdditiveSharingTensor


def test_wrap(workers):
    """
    Test the .on() wrap functionality for AdditiveSharingTensor
    """

    x_tensor = torch.Tensor([1, 2, 3])
    x = AdditiveSharingTensor().on(x_tensor)
    assert isinstance(x, torch.Tensor)
    assert isinstance(x.child, AdditiveSharingTensor)
    assert isinstance(x.child.child, torch.Tensor)


def test__str__(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])
    x_sh = th.tensor([[3, 4]]).share(alice, bob, crypto_provider=james)
    assert isinstance(x_sh.__str__(), str)


def test_encode_decode(workers):

    t = torch.tensor([1, 2, 3])
    x = t.share(workers["bob"], workers["alice"], workers["james"])

    x = x.get()

    assert (x == t).all()


def test_virtual_get(workers):
    t = torch.tensor([1, 2, 3])
    x = t.share(workers["bob"], workers["alice"], workers["james"])

    x = x.child.virtual_get()

    assert (x == t).all()


def test_autograd_kwarg(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])

    t = torch.tensor([1, 2, 3])
    x = t.share(alice, bob, crypto_provider=james, requires_grad=True)

    assert isinstance(x.child, syft.AutogradTensor)


def test_send_get(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])
    x_sh = th.tensor([[3, 4]]).share(alice, bob, crypto_provider=james)

    alice_t_id = x_sh.child.child["alice"].id_at_location
    assert alice_t_id in alice._objects

    ptr_x = x_sh.send(james)
    ptr_x_id_at_location = ptr_x.id_at_location
    assert ptr_x_id_at_location in james._objects
    assert alice_t_id in alice._objects

    x_sh_back = ptr_x.get()
    assert ptr_x_id_at_location not in james._objects
    assert alice_t_id in alice._objects

    x = x_sh_back.get()
    assert alice_t_id not in alice._objects


def test_add(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])

    # 2 workers
    t = torch.tensor([1, 2, 3])
    x = torch.tensor([1, 2, 3]).share(bob, alice)

    y = (x + x).get()

    # 3 workers
    assert (y == (t + t)).all()

    t = torch.tensor([1, 2, 3])
    x = torch.tensor([1, 2, 3]).share(bob, alice, james)

    y = (x + x).get()

    # negative numbers
    assert (y == (t + t)).all()

    t = torch.tensor([1, -2, 3])
    x = torch.tensor([1, -2, 3]).share(bob, alice, james)

    y = (x + x).get()

    assert (y == (t + t)).all()

    # with fixed precisions
    t = torch.tensor([1.0, -2, 3])
    x = torch.tensor([1.0, -2, 3]).fix_prec().share(bob, alice, james)

    y = (x + x).get().float_prec()

    assert (y == (t + t)).all()

    # with FPT>torch.tensor
    t = torch.tensor([1.0, -2.0, 3.0])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    y = t.fix_prec()

    z = (x + y).get().float_prec()

    assert (z == (t + t)).all()

    z = (y + x).get().float_prec()

    assert (z == (t + t)).all()


def test_sub(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])

    # 3 workers
    t = torch.tensor([1, 2, 3])
    x = torch.tensor([1, 2, 3]).share(bob, alice, james)

    y = (x - x).get()

    assert (y == (t - t)).all()

    # negative numbers
    t = torch.tensor([1, -2, 3])
    x = torch.tensor([1, -2, 3]).share(bob, alice, james)

    y = (x - x).get()

    assert (y == (t - t)).all()

    # with fixed precision
    t = torch.tensor([1.0, -2, 3])
    x = torch.tensor([1.0, -2, 3]).fix_prec().share(bob, alice, james)

    y = (x - x).get().float_prec()

    assert (y == (t - t)).all()

    # with FPT>torch.tensor
    t = torch.tensor([1.0, -2.0, 3.0])
    u = torch.tensor([4.0, 3.0, 2.0])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    y = u.fix_prec()

    z = (x - y).get().float_prec()

    assert (z == (t - u)).all()

    z = (y - x).get().float_prec()

    assert (z == (u - t)).all()


def test_mul(workers):
    torch.manual_seed(121)  # Truncation might not always work so we set the random seed
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])

    # 2 workers
    t = torch.tensor([1, 2, 3, 4])
    x = t.share(bob, alice, crypto_provider=james)
    y = (x * x).get()

    assert (y == (t * t)).all()

    # with fixed precision
    t = torch.tensor([1, 2, 3, 4.0])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    y = (x * x).get().float_prec()

    assert (y == (t * t)).all()

    # with non-default fixed precision
    t = torch.tensor([1, 2, 3, 4.0])
    x = t.fix_prec(precision_fractional=2).share(bob, alice, crypto_provider=james)
    y = (x * x).get().float_prec()

    assert (y == (t * t)).all()

    # with FPT>torch.tensor
    t = torch.tensor([1.0, -2.0, 3.0])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    y = t.fix_prec()

    z = (x * y).get().float_prec()

    assert (z == (t * t)).all()


def test_public_mul(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])

    t = th.tensor([-3.1, 1.0])
    x = t.fix_prec().share(alice, bob, crypto_provider=james)
    y = 1
    z = (x * y).get().float_prec()
    assert (z == (t * y)).all()

    t = th.tensor([-3.1, 1.0])
    x = t.fix_prec().share(alice, bob, crypto_provider=james)
    y = 0
    z = (x * y).get().float_prec()
    assert (z == (t * y)).all()

    t_x = th.tensor([-3.1, 1])
    t_y = th.tensor([1.0])
    x = t_x.fix_prec().share(alice, bob, crypto_provider=james)
    y = t_y.fix_prec()
    z = x * y
    z = z.get().float_prec()
    assert (z == t_x * t_y).all()

    t_x = th.tensor([-3.1, 1])
    t_y = th.tensor([0.0])
    x = t_x.fix_prec().share(alice, bob, crypto_provider=james)
    y = t_y.fix_prec()
    z = x * y
    z = z.get().float_prec()
    assert (z == t_x * t_y).all()

    t_x = th.tensor([-3.1, 1])
    t_y = th.tensor([0.0, 2.1])
    x = t_x.fix_prec().share(alice, bob, crypto_provider=james)
    y = t_y.fix_prec()
    z = x * y
    z = z.get().float_prec()
    assert (z == t_x * t_y).all()


def test_pow(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])

    m = torch.tensor([[1, 2], [3, 4.0]])
    x = m.fix_prec().share(bob, alice, crypto_provider=james)
    y = (x ** 3).get().float_prec()

    assert (y == (m ** 3)).all()


def test_operate_with_integer_constants(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])
    x = th.tensor([2.0])
    x_sh = x.fix_precision().share(alice, bob, crypto_provider=james)

    r_sh = x_sh + 10
    assert r_sh.get().float_prec() == x + 10

    r_sh = x_sh - 7
    assert r_sh.get().float_prec() == x - 7

    r_sh = x_sh * 2
    assert r_sh.get().float_prec() == x * 2

    r_sh = x_sh / 2
    assert r_sh.get().float_prec() == x / 2


def test_stack(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])
    t = torch.tensor([1.3, 2])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    res = torch.stack([x, x]).get().float_prec()

    expected = torch.tensor([[1.3000, 2.0000], [1.3000, 2.0000]])

    assert (res == expected).all()


def test_cat(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])
    t = torch.tensor([[1, 2, 3], [4, 5, 6]])
    x = t.share(bob, alice, crypto_provider=james)

    res0 = torch.cat([x, x], dim=0).get()
    res1 = torch.cat([x, x], dim=1).get()

    expected0 = torch.tensor([[1, 2, 3], [4, 5, 6], [1, 2, 3], [4, 5, 6]])
    expected1 = torch.tensor([[1, 2, 3, 1, 2, 3], [4, 5, 6, 4, 5, 6]])

    assert (res0 == expected0).all()
    assert (res1 == expected1).all()


def test_chunk(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])
    t = torch.tensor([[1, 2, 3, 4], [5, 6, 7, 8]])
    x = t.share(bob, alice, crypto_provider=james)

    res0 = torch.chunk(x, 2, dim=0)
    res1 = torch.chunk(x, 2, dim=1)

    expected0 = [torch.tensor([[1, 2, 3, 4]]), torch.tensor([[5, 6, 7, 8]])]
    expected1 = [torch.tensor([[1, 2], [5, 6]]), torch.tensor([[3, 4], [7, 8]])]

    assert all([(res0[i].get() == expected0[i]).all() for i in range(2)])
    assert all([(res1[i].get() == expected1[i]).all() for i in range(2)])


def test_roll(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])
    t = torch.tensor([[1, 2, 3], [4, 5, 6]])
    x = t.share(bob, alice, crypto_provider=james)

    res1 = torch.roll(x, 2)
    res2 = torch.roll(x, 2, dims=1)
    res3 = torch.roll(x, (1, 2), dims=(0, 1))

    assert (res1.get() == torch.roll(t, 2)).all()
    assert (res2.get() == torch.roll(t, 2, dims=1)).all()
    assert (res3.get() == torch.roll(t, (1, 2), dims=(0, 1))).all()

    # With MultiPointerTensor
    shifts = torch.tensor(1).send(alice, bob)
    res = torch.roll(x, shifts)

    shifts1 = torch.tensor(1).send(alice, bob)
    shifts2 = torch.tensor(2).send(alice, bob)
    res2 = torch.roll(x, (shifts1, shifts2), dims=(0, 1))

    assert (res.get() == torch.roll(t, 1)).all()
    assert (res2.get() == torch.roll(t, (1, 2), dims=(0, 1))).all()


def test_nn_linear(workers):
    torch.manual_seed(121)  # Truncation might not always work so we set the random seed
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])
    t = torch.tensor([[1.0, 2]])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    model = nn.Linear(2, 1)
    model.weight = nn.Parameter(torch.tensor([[-1.0, 2]]))
    model.bias = nn.Parameter(torch.tensor([[-1.0]]))
    model.fix_precision().share(bob, alice, crypto_provider=james)

    y = model(x)

    assert y.get().float_prec() == torch.tensor([[2.0]])


def test_matmul(workers):
    torch.manual_seed(121)  # Truncation might not always work so we set the random seed
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])

    m = torch.tensor([[1, 2], [3, 4.0]])
    x = m.fix_prec().share(bob, alice, crypto_provider=james)
    y = (x @ x).get().float_prec()

    assert (y == (m @ m)).all()

    # with FPT>torch.tensor
    m = torch.tensor([[1, 2], [3, 4.0]])
    x = m.fix_prec().share(bob, alice, crypto_provider=james)
    y = m.fix_prec()

    z = (x @ y).get().float_prec()

    assert (z == (m @ m)).all()

    z = (y @ x).get().float_prec()

    assert (z == (m @ m)).all()


def test_torch_conv2d(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])

    im = torch.Tensor(
        [
            [
                [[0.0, 1.0, 2.0], [3.0, 4.0, 5.0], [6.0, 7.0, 8.0]],
                [[10.0, 11.0, 12.0], [13.0, 14.0, 15.0], [16.0, 17.0, 18.0]],
            ]
        ]
    )
    w = torch.Tensor(
        [
            [[[1.0, 1.0], [1.0, 1.0]], [[2.0, 2.0], [2.0, 2.0]]],
            [[[-1.0, -2.0], [-3.0, -4.0]], [[0.0, 0.0], [0.0, 0.0]]],
        ]
    )
    bias = torch.Tensor([0.0, 5.0])

    im_shared = im.fix_precision().share(bob, alice, crypto_provider=james)
    w_shared = w.fix_precision().share(bob, alice, crypto_provider=james)
    bias_shared = bias.fix_precision().share(bob, alice, crypto_provider=james)

    res0 = torch.conv2d(im_shared, w_shared, bias=bias_shared, stride=1).get().float_precision()
    res1 = (
        torch.conv2d(
            im_shared,
            w_shared[:, 0:1].contiguous(),
            bias=bias_shared,
            stride=2,
            padding=3,
            dilation=2,
            groups=2,
        )
        .get()
        .float_precision()
    )

    expected0 = torch.conv2d(im, w, bias=bias, stride=1)
    expected1 = torch.conv2d(
        im, w[:, 0:1].contiguous(), bias=bias, stride=2, padding=3, dilation=2, groups=2
    )

    assert (res0 == expected0).all()
    assert (res1 == expected1).all()


def test_fixed_precision_and_sharing(workers):

    bob, alice = (workers["bob"], workers["alice"])

    t = torch.tensor([1, 2, 3, 4.0])
    x = t.fix_prec().share(bob, alice)
    out = x.get().float_prec()

    assert (out == t).all()

    x = t.fix_prec().share(bob, alice)

    y = x + x

    y = y.get().float_prec()
    assert (y == (t + t)).all()


def test_fixed_precision_and_sharing_on_pointer(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])

    t = torch.tensor([1, 2, 3, 4.0])
    ptr = t.send(james)

    x = ptr.fix_prec().share(bob, alice)

    y = x + x

    y = y.get().get().float_prec()
    assert (y == (t + t)).all()


def test_pointer_on_fixed_precision_and_sharing(workers):
    bob, alice, james = (workers["bob"], workers["alice"], workers["james"])

    t = torch.tensor([1, 2, 3, 4.0])

    x = t.fix_prec().share(bob, alice)
    x = x.send(james)

    y = x + x

    y = y.get().get().float_prec()
    assert (y == (t + t)).all()


def test_get_item(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]
    x = th.tensor([[3.1, 4.3]]).fix_prec().share(alice, bob, crypto_provider=james)
    idx = torch.tensor([0]).send(alice, bob)

    # Operate directly AST[MPT]
    assert x.child.child[:, idx.child].get() == torch.tensor([[3100]])

    # With usual wrappers and FPT
    x = th.tensor([[3, 4]]).share(alice, bob, crypto_provider=james)
    idx = torch.tensor([0]).send(alice, bob)
    assert x[:, idx].get() == torch.tensor([[3]])


def test_eq(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    x = th.tensor([3.1]).fix_prec().share(alice, bob, crypto_provider=james)
    y = th.tensor([3.1]).fix_prec().share(alice, bob, crypto_provider=james)

    assert (x == y).get().float_prec()

    x = th.tensor([3.1]).fix_prec().share(alice, bob, crypto_provider=james)
    y = th.tensor([2.1]).fix_prec().share(alice, bob, crypto_provider=james)

    assert not (x == y).get().float_prec()

    x = th.tensor([-3.1]).fix_prec().share(alice, bob, crypto_provider=james)
    y = th.tensor([-3.1]).fix_prec().share(alice, bob, crypto_provider=james)

    assert (x == y).get().float_prec()


def test_comp(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    x = th.tensor([3.1]).fix_prec().share(alice, bob, crypto_provider=james)
    y = th.tensor([3.1]).fix_prec().share(alice, bob, crypto_provider=james)

    assert (x >= y).get().float_prec()
    assert (x <= y).get().float_prec()
    assert not (x > y).get().float_prec()
    assert not (x < y).get().float_prec()

    x = th.tensor([-3.1]).fix_prec().share(alice, bob, crypto_provider=james)
    y = th.tensor([-3.1]).fix_prec().share(alice, bob, crypto_provider=james)

    assert (x >= y).get().float_prec()
    assert (x <= y).get().float_prec()
    assert not (x > y).get().float_prec()
    assert not (x < y).get().float_prec()

    x = th.tensor([3.1]).fix_prec().share(alice, bob, crypto_provider=james)
    y = th.tensor([2.1]).fix_prec().share(alice, bob, crypto_provider=james)

    assert (x >= y).get().float_prec()
    assert not (x <= y).get().float_prec()
    assert (x > y).get().float_prec()
    assert not (x < y).get().float_prec()

    x = th.tensor([-2.1]).fix_prec().share(alice, bob, crypto_provider=james)
    y = th.tensor([-3.1]).fix_prec().share(alice, bob, crypto_provider=james)

    assert (x >= y).get().float_prec()
    assert not (x <= y).get().float_prec()
    assert (x > y).get().float_prec()
    assert not (x < y).get().float_prec()


def test_max(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    t = torch.tensor([3, 1.0, 2])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    max_value = x.max().get().float_prec()
    assert max_value == torch.tensor([3.0])

    t = torch.tensor([3, 4.0])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    max_value = x.max().get().float_prec()
    assert max_value == torch.tensor([4.0])

    t = torch.tensor([3, 4.0, 5, 2])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    max_value = x.max().get().float_prec()
    assert max_value == torch.tensor([5.0])


def test_argmax(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    t = torch.tensor([3, 1.0, 2])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    idx = x.argmax().get().float_prec()
    assert idx == torch.tensor([0.0])

    t = torch.tensor([3, 4.0])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    idx = x.argmax().get().float_prec()
    assert idx == torch.tensor([1.0])

    t = torch.tensor([3, 4.0, 5, 2])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    idx = x.argmax().get().float_prec()
    assert idx == torch.tensor([2.0])

    # no dim=
    t = torch.tensor([[1, 2.0, 4], [3, 9.0, 2.0]])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    ids = x.argmax().get().float_prec()
    assert ids.long() == torch.argmax(t)  # TODO rm .long()

    # dim=1
    t = torch.tensor([[1, 2.0, 4], [3, 1.0, 2.0]])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    ids = x.argmax(dim=1).get().float_prec()
    assert (ids.long() == torch.argmax(t, dim=1)).all()  # TODO rm .long()


def test_mod(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    t = torch.tensor([21]).share(bob, alice, crypto_provider=james)
    assert t.child.mod(8).get() % 8 == torch.tensor([5])
    assert t.child.mod(-8).get() % -8 == torch.tensor([-3])

    t = torch.tensor([-21]).share(bob, alice, crypto_provider=james)
    assert t.child.mod(8).get() % 8 == torch.tensor([3])
    assert t.child.mod(-8).get() % -8 == torch.tensor([-5])

    assert (t.child % 8).get() % 8 == torch.tensor([3])


def test_torch_sum(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    t = torch.tensor([[1, 2, 4], [8, 5, 6]])
    x = t.share(alice, bob, crypto_provider=james)

    s = torch.sum(x).get()
    s_dim = torch.sum(x, 0).get()
    s_dim2 = torch.sum(x, (0, 1)).get()
    s_keepdim = torch.sum(x, 1, keepdim=True).get()

    assert (s == torch.sum(t)).all()
    assert (s_dim == torch.sum(t, 0)).all()
    assert (s_dim2 == torch.sum(t, (0, 1))).all()
    assert (s_keepdim == torch.sum(t, 1, keepdim=True)).all()


def test_torch_mean(workers):
    torch.manual_seed(121)  # Truncation might not always work so we set the random seed
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]
    base = 10
    prec_frac = 4

    t = torch.tensor([[1.0, 2.5], [8.0, 5.5]])
    x = t.fix_prec(base=base, precision_fractional=prec_frac).share(
        alice, bob, crypto_provider=james
    )

    s = torch.mean(x).get().float_prec()
    s_dim = torch.mean(x, 0).get().float_prec()
    s_dim2 = torch.mean(x, (0, 1)).get().float_prec()
    s_keepdim = torch.mean(x, 1, keepdim=True).get().float_prec()

    assert (s == torch.tensor(4.25)).all()
    assert (s_dim == torch.tensor([4.5, 4.0])).all()
    assert (s_dim2 == torch.tensor(4.25)).all()
    assert (s_keepdim == torch.tensor([[1.75], [6.75]])).all()


def test_torch_dot(workers):
    torch.manual_seed(121)  # Truncation might not always work so we set the random seed
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    x = torch.tensor([1.0, 2.0, 3.0, 4.0, 5.0]).fix_prec().share(alice, bob, crypto_provider=james)
    y = torch.tensor([3.0, 3.0, 3.0, 3.0, 3.0]).fix_prec().share(alice, bob, crypto_provider=james)

    assert torch.dot(x, y).get().float_prec() == 45


def test_unbind(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    x = torch.tensor([21, 17]).share(bob, alice, crypto_provider=james).child

    x0, x1 = torch.unbind(x)

    assert x0.get() == torch.tensor(21)
    assert x1.get() == torch.tensor(17)


def test_handle_func_command(workers):
    """
    Just to show that handle_func_command works
    Even if torch.abs should be hooked to return a correct value
    """
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    t = torch.tensor([-21]).share(bob, alice, crypto_provider=james).child
    _ = torch.abs(t).get()


def test_init_with_no_crypto_provider(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    x = torch.tensor([21, 17]).share(bob, alice).child

    assert x.crypto_provider.id == syft.hook.local_worker.id


def test_zero_refresh(workers):
    alice, bob, james = workers["alice"], workers["bob"], workers["james"]

    t = torch.tensor([2.2, -1.0])
    x = t.fix_prec().share(bob, alice, crypto_provider=james)

    x_sh = x.child.child
    assert (x_sh.zero().get() == torch.zeros(*t.shape).long()).all()

    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    x_copy = t.fix_prec().share(bob, alice, crypto_provider=james)
    x_r = x.refresh()

    assert (x_r.get().float_prec() == x_copy.get().float_prec()).all()

    x = t.fix_prec().share(bob, alice, crypto_provider=james)
    x_r = x.refresh()

    assert ((x_r / 2).get().float_prec() == t / 2).all()



"""Specific Pysyft exceptions."""

import syft as sy
import torch
from tblib import Traceback
import traceback
from six import reraise
from typing import Tuple


class PureTorchTensorFoundError(BaseException):
    """Exception raised for errors in the input.
    This error is used in a recursive analysis of the args provided as an
    input of a function, to break the recursion if a TorchTensor is found
    as it means that _probably_ all the tensors are pure torch tensor and
    the function can be applied natively on this input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    pass


class RemoteObjectFoundError(BaseException):
    """Exception raised for errors in the input.
    This error is used in a context similar to PureTorchTensorFoundError but
    to indicate that a Pointer to a remote tensor was found  in the input
    and thus that the command should be send elsewhere. The pointer retrieved
    by the error gives the location where the command should be sent.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, pointer):
        self.pointer = pointer


class InvalidTensorForRemoteGet(Exception):
    """Raised when a chain of pointer tensors is not provided for `remote_get`."""

    def __init__(self, tensor: object):
        message = "Tensor does not have attribute child. You remote get should be called on a chain of pointer tensors, instead you called it on {}.".format(
            tensor
        )
        super().__init__(message)


class WorkerNotFoundException(Exception):
    """Raised when a non-existent worker is requested."""

    pass


class CompressionNotFoundException(Exception):
    """Raised when a non existent compression/decompression scheme is requested."""

    pass


class CannotRequestObjectAttribute(Exception):
    """Raised when .get() is called on a pointer which points to an attribute of
    another object."""

    pass


class TensorsNotCollocatedException(Exception):
    """Raised when a command is executed on two tensors which are not
    on the same machine. The goal is to provide as useful input as possible
    to help the user identify which tensors are where so that they can debug
    which one needs to be moved."""

    def __init__(self, tensor_a, tensor_b, attr="a method"):

        if hasattr(tensor_a, "child") and tensor_a.is_wrapper:
            tensor_a = tensor_a.child

        if hasattr(tensor_b, "child") and tensor_b.is_wrapper:
            tensor_b = tensor_b.child

        if isinstance(tensor_a, sy.PointerTensor) and isinstance(tensor_b, sy.PointerTensor):
            message = (
                "You tried to call "
                + attr
                + " involving two tensors which"
                + " are not on the same machine! One tensor is on "
                + str(tensor_a.location)
                + " while the other is on "
                + str(tensor_b.location)
                + ". Use a combination of .move(), .get(), and/or .send() to co-locate them to the same machine."
            )
        elif isinstance(tensor_a, sy.PointerTensor):
            message = (
                "You tried to call "
                + attr
                + " involving two tensors where one tensor is actually located"
                + " on another machine (is a PointerTensor). Call .get() on the PointerTensor or .send("
                + str(tensor_a.location.id)
                + ") on the other tensor.\n"
                + "\nTensor A: "
                + str(tensor_a)
                + "\nTensor B: "
                + str(tensor_b)
            )
        elif isinstance(tensor_b, sy.PointerTensor):
            message = (
                "You tried to call "
                + attr
                + " involving two tensors where one tensor is actually located"
                + " on another machine (is a PointerTensor). Call .get() on the PointerTensor or .send("
                + str(tensor_b.location.id)
                + ") on the other tensor.\n"
                + "\nTensor A: "
                + str(tensor_a)
                + "\nTensor B: "
                + str(tensor_b)
            )
        else:
            message = (
                "You tried to call "
                + attr
                + " involving two tensors which are not on the same machine."
                + "Try calling .send(), .move(), and/or .get() on these tensors to get them to the same"
                + "worker before calling methods that involve them working together."
            )

        super().__init__(message)

        self.tensor_a = tensor_a
        self.tensor_b = tensor_b


class ResponseSignatureError(Exception):
    """Raised when the return of a hooked function is not correctly predicted
    (when defining in advance ids for results)"""

    def __init__(self, ids_generated=None):
        self.ids_generated = ids_generated

    def get_attributes(self):
        """
        Specify all the attributes need to report an error correctly.
        """
        return {"ids_generated": self.ids_generated}

    @staticmethod
    def simplify(e):
        """
        Serialize information about an Exception which was raised to forward it
        """
        # Get information about the exception: type of error,  traceback
        tp = type(e)
        tb = e.__traceback__
        # Serialize the traceback
        traceback_str = "Traceback (most recent call last):\n" + "".join(traceback.format_tb(tb))
        # Include special attributes if relevant
        try:
            attributes = e.get_attributes()
        except AttributeError:
            attributes = {}
        return tp.__name__, traceback_str, sy.serde._simplify(attributes)

    @staticmethod
    def detail(worker: "sy.workers.AbstractWorker", error_tuple: Tuple[str, str, dict]):
        """
        Detail and re-raise an Exception forwarded by another worker
        """
        error_name, traceback_str, attributes = error_tuple
        error_name, traceback_str = error_name.decode("utf-8"), traceback_str.decode("utf-8")
        attributes = sy.serde._detail(worker, attributes)
        # De-serialize the traceback
        tb = Traceback.from_string(traceback_str)
        # Check that the error belongs to a valid set of Exceptions
        if error_name in dir(sy.exceptions):
            error_type = getattr(sy.exceptions, error_name)
            error = error_type()
            # Include special attributes if any
            for attr_name, attr in attributes.items():
                setattr(error, attr_name, attr)
            reraise(error_type, error, tb.as_traceback())
        else:
            raise ValueError(f"Invalid Exception returned:\n{traceback_str}")


class GetNotPermittedError(Exception):
    """Raised when calling get on a pointer to a tensor which does not allow
    get to be called on it. This can happen do to sensitivity being too high"""

    @staticmethod
    def simplify(e):
        """
        Serialize information about an Exception which was raised to forward it
        """
        # Get information about the exception: type of error,  traceback
        tp = type(e)
        tb = e.__traceback__
        # Serialize the traceback
        traceback_str = "Traceback (most recent call last):\n" + "".join(traceback.format_tb(tb))
        # Include special attributes if relevant
        try:
            attributes = e.get_attributes()
        except AttributeError:
            attributes = {}
        return tp.__name__, traceback_str, sy.serde._simplify(attributes)

    @staticmethod
    def detail(worker: "sy.workers.AbstractWorker", error_tuple: Tuple[str, str, dict]):
        """
        Detail and re-raise an Exception forwarded by another worker
        """
        error_name, traceback_str, attributes = error_tuple
        error_name, traceback_str = error_name.decode("utf-8"), traceback_str.decode("utf-8")
        attributes = sy.serde._detail(worker, attributes)
        # De-serialize the traceback
        tb = Traceback.from_string(traceback_str)
        # Check that the error belongs to a valid set of Exceptions
        if error_name in dir(sy.exceptions):
            error_type = getattr(sy.exceptions, error_name)
            error = error_type()
            # Include special attributes if any
            for attr_name, attr in attributes.items():
                setattr(error, attr_name, attr)
            reraise(error_type, error, tb.as_traceback())
        else:
            raise ValueError(f"Invalid Exception returned:\n{traceback_str}")


class IdNotUniqueError(Exception):
    """Raised by the ID Provider when setting ids that have already been generated"""

    pass


def route_method_exception(exception, self, args, kwargs):
    try:
        if self.is_wrapper:
            if isinstance(self.child, sy.PointerTensor):
                if len(args) > 0:
                    if not args[0].is_wrapper:
                        return TensorsNotCollocatedException(self, args[0])
                    elif isinstance(args[0].child, sy.PointerTensor):
                        if self.location != args[0].child.location:
                            return TensorsNotCollocatedException(self, args[0])

        # if self is a normal tensor
        elif isinstance(self, torch.Tensor):
            if len(args) > 0:
                if args[0].is_wrapper:
                    if isinstance(args[0].child, sy.PointerTensor):
                        return TensorsNotCollocatedException(self, args[0])
                elif isinstance(args[0], sy.PointerTensor):
                    return TensorsNotCollocatedException(self, args[0])
    except:
        ""
    return exception



import torch
from typing import Callable
import syft as sy
from syft.workers.abstract import AbstractWorker

no_wrap = {"no_wrap": True}


def spdz_mul(cmd: Callable, x_sh, y_sh, crypto_provider: AbstractWorker, field: int):
    """Abstractly multiplies two tensors (mul or matmul)

    Args:
        cmd: a callable of the equation to be computed (mul or matmul)
        x_sh (AdditiveSharingTensor): the left part of the operation
        y_sh (AdditiveSharingTensor): the right part of the operation
        crypto_provider (AbstractWorker): an AbstractWorker which is used to generate triples
        field (int): an integer denoting the size of the field

    Return:
        an AdditiveSharingTensor
    """
    assert isinstance(x_sh, sy.AdditiveSharingTensor)
    assert isinstance(y_sh, sy.AdditiveSharingTensor)

    locations = x_sh.locations

    # Get triples
    a, b, a_mul_b = crypto_provider.generate_triple(cmd, field, x_sh.shape, y_sh.shape, locations)

    delta = x_sh - a
    epsilon = y_sh - b
    # Reconstruct and send to all workers
    delta = delta.reconstruct()
    epsilon = epsilon.reconstruct()

    delta_epsilon = cmd(delta, epsilon)

    # Trick to keep only one child in the MultiPointerTensor (like in SNN)
    j1 = torch.ones(delta_epsilon.shape).long().send(locations[0], **no_wrap)
    j0 = torch.zeros(delta_epsilon.shape).long().send(*locations[1:], **no_wrap)
    if len(locations) == 2:
        j = sy.MultiPointerTensor(children=[j1, j0])
    else:
        j = sy.MultiPointerTensor(children=[j1] + j0.child.values())

    delta_b = cmd(delta, b)
    a_epsilon = cmd(a, epsilon)

    return delta_epsilon * j + delta_b + a_epsilon + a_mul_b
